@@ -22,7 +22,7 @@ import fortranformat
 from omfit_classes.utils_fusion import tokamak
 import omas
 
-__all__ = ['OMFITmhdin', 'OMFITdprobe', 'OMFITnstxMHD', 'get_mhdindat']
+__all__ = ['OMFITmhdin', 'OMFITdprobe', 'OMFITnstxMHD', 'get_mhdindat', 'green_to_omas']
 
 
 class OMFITmhdin(OMFITnamelist):
@@ -46,6 +46,7 @@ class OMFITmhdin(OMFITnamelist):
         if 'IN3' not in self:
             self['IN3'] = NamelistName()
         if 'MACHINEIN' not in self:
+
             self['MACHINEIN'] = NamelistName()
         # out-of-namelist format (DIII-D)
 
@@ -280,7 +281,7 @@ class OMFITmhdin(OMFITnamelist):
             ax.scatter(x0, y0, s=s0, color='b', marker='o', alpha=0.75, zorder=100)
 
         # labels
-        if label:
+        if label and 'LPNAME' in self['IN3']:
             for k, name in enumerate(self['IN3']['LPNAME']):
                 ax.text(x0[k], y0[k], '\n ' + name, ha='left', va='top', size=8, color='w', zorder=1000, weight='bold', clip_on=True)
                 ax.text(x0[k], y0[k], '\n ' + name, ha='left', va='top', size=8, color='b', zorder=1001, clip_on=True)
@@ -394,7 +395,7 @@ class OMFITmhdin(OMFITnamelist):
                 ax.plot(segx[k], segy[k], '.-', lw=2, color=col, zorder=100, alpha=0.75, mec='none')
 
         # labels
-        if label:
+        if label and 'MPNAM2' in self['IN3']:
             for k, name in enumerate(self['IN3']['MPNAM2']):
                 ax.text(x0[k], y0[k], '\n ' + name, ha='left', va='top', size=8, color='w', zorder=1000, weight='bold', clip_on=True)
                 ax.text(x0[k], y0[k], '\n ' + name, ha='left', va='top', size=8, color='r', zorder=1001, clip_on=True)
@@ -474,6 +475,8 @@ class OMFITmhdin(OMFITnamelist):
             system_array = np.array(list(zip(in3[f'R{bn}'], in3[f'Z{bn}'], in3[f'W{bn}'], in3[f'H{bn}'], in3[f'A{bn}'], in3[f'A{bn}2'])))
 
         if system == 'OH':
+            if len(in3['RE']) == 0:
+                return
             n = int(max(system_array[:, -1]))
         for k in range(system_array.shape[0]):
             # disable plotting of dummy probes/loops
@@ -770,7 +773,7 @@ class OMFITmhdin(OMFITnamelist):
     def efund_to_outline(self, coil_data, outline):
 
         """
-        The routine calculates efund data format to ods outline format
+        The routine converts efund data format to ods outline format
 
          :param coil_data: 6-index array, r,z,w,h,a1,a2
 
@@ -779,28 +782,54 @@ class OMFITmhdin(OMFITnamelist):
          :return: outline
         """
 
-        fdat = coil_data.copy()
-        fdat[4] = -coil_data[4] * np.pi / 180.0
-        fdat[5] = -(coil_data[5] * np.pi / 180.0 if coil_data[5] != 0 else np.pi / 2)
-        outline['r'] = [
-            fdat[0] - fdat[2] / 2.0 - fdat[3] / 2.0 * np.tan((np.pi / 2.0 + fdat[5])),
-            fdat[0] - fdat[2] / 2.0 + fdat[3] / 2.0 * np.tan((np.pi / 2.0 + fdat[5])),
-            fdat[0] + fdat[2] / 2.0 + fdat[3] / 2.0 * np.tan((np.pi / 2.0 + fdat[5])),
-            fdat[0] + fdat[2] / 2.0 - fdat[3] / 2.0 * np.tan((np.pi / 2.0 + fdat[5])),
-        ]
+        a1 = coil_data[4] * np.pi / 180.0
+        a2 = coil_data[5] * np.pi / 180.0
+        if a1 < 1e-8:
+            side = coil_data[3] / np.tan(a2)
+            hw1 = (coil_data[2] + side) / 2.0
+            hw2 = (coil_data[2] - side) / 2.0
+            hh = coil_data[3] / 2.0
+            outline['r'] = [
+                coil_data[0] - hw1,
+                coil_data[0] - hw2,
+                coil_data[0] + hw1,
+                coil_data[0] + hw2,
+            ]
+            outline['z'] = [
+                coil_data[1] - hh,
+                coil_data[1] + hh,
+                coil_data[1] + hh,
+                coil_data[1] - hh,
+            ]
+        else:
+            side = coil_data[2] * np.tan(a1)
+            hw = coil_data[2] / 2.0
+            hh1 = (coil_data[3] + side) / 2.0
+            hh2 = (coil_data[3] - side) / 2.0
+            outline['r'] = [
+                coil_data[0] - hw,
+                coil_data[0] - hw,
+                coil_data[0] + hw,
+                coil_data[0] + hw,
+            ]
+            outline['z'] = [
+                coil_data[1] - hh1,
+                coil_data[1] + hh2,
+                coil_data[1] + hh1,
+                coil_data[1] - hh2,
+            ]
 
-        outline['z'] = [
-            fdat[1] - fdat[3] / 2.0 - fdat[2] / 2.0 * np.tan(-fdat[4]),
-            fdat[1] + fdat[3] / 2.0 - fdat[2] / 2.0 * np.tan(-fdat[4]),
-            fdat[1] + fdat[3] / 2.0 + fdat[2] / 2.0 * np.tan(-fdat[4]),
-            fdat[1] - fdat[3] / 2.0 + fdat[2] / 2.0 * np.tan(-fdat[4]),
-        ]
         return outline
 
     def outline_to_efund(self, outline):
 
         """
-        The routine calculates ods outline format to efund data format
+        The routine converts ods outline format to efund data format
+          Since efund only supports parallelograms and requires 2 sides
+            to be either vertical or horizontal this will likely not match
+            the outline very well.  Instead, the parallelogram will only
+            match the angle of the lower left side, the height of the upper
+            right side, and width of the the left most top side.
 
          :param outline: ods outline entry
 
@@ -812,7 +841,7 @@ class OMFITmhdin(OMFITnamelist):
         rrel = outline['r'] - rcent
         zrel = outline['z'] - zcent
 
-        angle = arctan2(zrel, rrel) % (2 * pi)
+        angle = np.arctan2(zrel, rrel)
         rsort = [r for _, r in sorted(zip(angle, rrel))]
         zsort = [z for _, z in sorted(zip(angle, zrel))]
 
@@ -821,14 +850,33 @@ class OMFITmhdin(OMFITnamelist):
         r00, z00 = rsort[2], zsort[2]
         r10, z10 = rsort[3], zsort[3]
 
-        a1 = arctan((z10 - z00) / (r10 - r00))
-        atmp = arctan((z11 - z10) / (r11 - r10))
-        a2 = atmp - a1
+        # handle fringe cases that should be valid
+        n = 0
+        for i in range(4):
+            if angle[i] <= -np.pi / 2:
+                n += 1
+        if n == 0:
+            if r11**2 + z11**2 > r10**2 + z10**2:
+                r11, z11 = rsort[3], zsort[3]
+                r01, z01 = rsort[0], zsort[0]
+                r00, z00 = rsort[1], zsort[1]
+                r10, z10 = rsort[2], zsort[2]
+        elif n == 2:
+            if r11**2 + z11**2 < r01**2 + z01**2:
+                r11, z11 = rsort[1], zsort[1]
+                r01, z01 = rsort[2], zsort[2]
+                r00, z00 = rsort[3], zsort[3]
+                r10, z10 = rsort[0], zsort[0]
 
-        wid = (r11 - r01) * cos(a1)
-        height = (z11 - z10) * sin(a1 + a2)
+        height = z00 - z01
+        width = r00 - r10
+        a1 = np.arctan2(z01 - z11, r01 - r11)
+        a2 = 0.0
+        if abs(a1) < 1.0e-8:
+            a1 = 0.0
+            a2 = np.arctan2(z00 - z01, r00 - r01)
 
-        return [rcent, zcent, wid, height, scipy.degrees(a1), scipy.degrees(a2)]
+        return [rcent, zcent, width, height, scipy.degrees(a1), scipy.degrees(a2)]
 
     def rectangle_to_efund(self, rectangle):
         r = rectangle['r']
@@ -839,6 +887,84 @@ class OMFITmhdin(OMFITnamelist):
         a2 = 0.0
         return [r, z, w, h, a1, a2]
 
+    def annulus_to_efund(self, annulus):
+        """
+        The routine converts an ods annulus format to efund data format
+          by approximating it as a square
+
+         :param annulus: ods annulus entry
+
+         :return: 6-index array, r,z,w,h,a1,a2
+        """
+        r = annulus['r']
+        z = annulus['z']
+        w = 2 * annulus['radius_outer']
+        h = w
+        a1 = 0.0
+        a2 = 0.0
+        return [r, z, w, h, a1, a2]
+
+    def thick_line_to_efund(self, thick_line):
+
+        """
+        The routine converts ods thick_line format to efund data format
+          The only time a thick line is a valid shape in efund is when
+            it is vertical or horizontal.  All others will not be a
+            great fit, but some approximation is used.
+
+         :param thick_line: ods thick_line entry
+
+         :return: 6-index array, r,z,w,h,a1,a2
+        """
+
+        r1, z1 = thick_line['first_point']['r'], thick_line['first_point']['z']
+        r2, z2 = thick_line['second_point']['r'], thick_line['second_point']['z']
+        a = np.pi / 2 - np.arctan2(z2 - z1, r2 - r1)
+        dr = thick_line['thickness'] / 2.0 * np.cos(a)
+        dz = thick_line['thickness'] / 2.0 * np.sin(a)
+
+        r11, z11 = r1 - dr, z1 + dz
+        r01, z01 = r1 + dr, z1 - dz
+        r00, z00 = r2 - dr, z2 + dz
+        r10, z10 = r2 + dr, z2 - dz
+        outline = {"r": [r11, r10, r00, r01], "z": [z11, z10, z00, z01]}
+
+        return self.outline_to_efund(outline)
+
+    def annular_to_efund(self, annular):
+
+        """
+        The routine converts ods annular format to efund data format
+          The only time annular segments are a valid shape in efund is when
+            they are vertical or horizontal.  All others will not be a
+            great fit, but some approximation is used.
+
+         :param annular: ods annular entry
+
+         :return: 6-index array, r,z,w,h,a1,a2 in which each is an array over
+                  the number of segments
+        """
+
+        ns = len(annular['centreline.r']) - 1
+        nt = ns + annular['centreline.closed']
+        r, z = np.zeros(nt), np.zeros(nt)
+        w, h = np.zeros(nt), np.zeros(nt)
+        a1, a2 = np.zeros(nt), np.zeros(nt)
+
+        for i in range(ns):
+            fp = {"r": annular['centreline.r'][i], "z": annular['centreline.z'][i]}
+            sp = {"r": annular['centreline.r'][i + 1], "z": annular['centreline.z'][i + 1]}
+            thick_line = {"first_point": fp, "second_point": sp, "thickness": annular['thickness'][i]}
+            [r[i], z[i], w[i], h[i], a1[i], a2[i]] = self.thick_line_to_efund(thick_line)
+
+        if annular['centreline.closed']:
+            fp = {"r": annular['centreline.r'][-1], "z": annular['centreline.z'][-1]}
+            sp = {"r": annular['centreline.r'][0], "z": annular['centreline.z'][0]}
+            thick_line = {"first_point": fp, "second_point": sp, "thickness": annular['thickness'][-1]}
+            [r[-1], z[-1], w[-1], h[-1], a1[-1], a2[-1]] = self.thick_line_to_efund(thick_line)
+
+        return [r, z, w, h, a1, a2]
+
     # Genirate a mhdin starting
     def init_mhdin(self, device):
         # mhdin = OMFITnamelist(filename = 'mhdin')
@@ -867,16 +993,14 @@ class OMFITmhdin(OMFITnamelist):
         self['IN3']['NSMP2'] = 25
         self['IN3']['IECOIL'] = 0
         self['IN3']['IVESEL'] = 0
-        self['IN3']['vsid'] = 1
         self['IN3']['IACOIL'] = 0
-        self['IN3']['nw'] = 129
-        self['IN3']['nh'] = 129
+        self['IN3']['NW'] = 129
+        self['IN3']['NH'] = 129
 
         self['IN3']['RF'] = []
         self['IN3']['ZF'] = []
         self['IN3']['WF'] = []
         self['IN3']['HF'] = []
-
         self['IN3']['AF'] = []
         self['IN3']['AF2'] = []
         self['IN3']['TURNFC'] = []
@@ -896,18 +1020,23 @@ class OMFITmhdin(OMFITnamelist):
         self['IN3']['HVS'] = []
         self['IN3']['AVS'] = []
         self['IN3']['AVS2'] = []
+        self['IN3']['RSISVS'] = []
+        self['IN3']['VSID'] = []
 
-        self['IN3']['RA'] = []
-        self['IN3']['ZA'] = []
-        self['IN3']['WA'] = []
-        self['IN3']['HA'] = []
+        self['IN3']['RACOIL'] = []
+        self['IN3']['ZACOIL'] = []
+        self['IN3']['WACOIL'] = []
+        self['IN3']['HACOIl'] = []
 
         self['IN3']['RSI'] = []
-
         self['IN3']['ZSI'] = []
+        #        self['IN3']['WSI'] = []
+        #        self['IN3']['HSI'] = []
+        #        self['IN3']['AS'] = []
+        #        self['IN3']['AS2'] = []
         return self
 
-    def from_omas(self, ods):
+    def from_omas(self, ods, passive_map='F'):
 
         if 'dataset_description.data_entry.machine' in ods:
             device = ods['dataset_description.data_entry.machine']
@@ -916,78 +1045,161 @@ class OMFITmhdin(OMFITnamelist):
 
         self = self.init_mhdin(device)
 
-        ncoil = {'F': 0, 'E': 0, 'A': 0}
-        coil_map = {'OH': 'E', 'PF': 'F', 'DIV': 'A'}
+        ncoil = {'F': 0, 'E': 0, 'ACOIL': 0, 'VS': 0}
+        nsum = {'F': 0, 'E': 0, 'ACOIL': 0, 'VS': 0}
+        # DIII-D
+        coil_map = {'OH': 'E', 'PF': 'F', 'DIV': 'ACOIL'}
+        # ITER
+        coil_map['CS'] = 'F'  # Central solenoid
+        # coil_map['CS'] = 'E' # probably a better choice once operational values are known
+        coil_map['VS'] = passive_map  # Vertical Stabilzation (in-vessel coils)
+        coil_map['TF'] = 'ACOIL'  # TF coil busbars
+        coil_map['VC'] = 'S'  # Virtual coils (skip)
 
-        for icoil in ods['pf_active']['coil']:
-            coil = ods['pf_active']['coil'][icoil]
-            for i in coil_map.keys():
-                if i in coil['name']:
-                    efund_name = coil_map[i]
+        for coil_type in ['pf_active.coil', 'pf_passive.loop']:
 
-            ncoil[efund_name] += 1
-            if 'rectangle' in coil['element'][0]['geometry']:
-                [r, z, w, h, a1, a2] = self.rectangle_to_efund(coil['element'][0]['geometry']['rectangle'])
-            else:
-                [r, z, w, h, a1, a2] = self.outline_to_efund(coil['element'][0]['geometry.outline'])
+            for isum in ods[coil_type]:
 
-            self['IN3'][f'R{efund_name}'].append(r)
-            self['IN3'][f'Z{efund_name}'].append(z)
-            self['IN3'][f'W{efund_name}'].append(w)
-            self['IN3'][f'H{efund_name}'].append(h)
+                coil = ods[coil_type][isum]
+                for i in coil_map.keys():
+                    if i in coil['name']:
+                        efund_name = coil_map[i]
 
-            if efund_name == 'F':
-                self['IN3'][f'A{efund_name}'].append(a1)
-                self['IN3'][f'A{efund_name}2'].append(a2)
+                if efund_name == 'S':
+                    continue
+                elif efund_name == 'VS' and coil_type == 'pf_passive.loop':
+                    self['IN3']['RSISVS'].append(coil['resistivity'])
 
-                self['IN3'][f'TURN{efund_name}C'].append(1.0)
+                nsum[efund_name] += 1
+                for ielement in coil['element']:
+                    ncoil[efund_name] += 1
+                    element = coil['element'][ielement]
+                    if 'rectangle' in element['geometry']:
+                        [r, z, w, h, a1, a2] = self.rectangle_to_efund(element['geometry.rectangle'])
+                    elif 'annulus' in element['geometry']:
+                        [r, z, w, h, a1, a2] = self.annulus_to_efund(element['geometry.annulus'])
+                    elif 'outline' in element['geometry']:
+                        [r, z, w, h, a1, a2] = self.outline_to_efund(element['geometry.outline'])
+                    else:
+                        raise ValueError(f'No conversion defined for geometry of coil {isum}')
 
-            if efund_name != 'A':
-                self['IN3'][f'{efund_name}CID'].append(len(self['IN3'][f'R{efund_name}']))
+                    self['IN3'][f'R{efund_name}'].append(r)
+                    self['IN3'][f'Z{efund_name}'].append(z)
+                    self['IN3'][f'W{efund_name}'].append(w)
+                    self['IN3'][f'H{efund_name}'].append(h)
 
-                self['IN3'][f'{efund_name}CTURN'].append(coil['element'][0]['turns_with_sign'])
+                    if efund_name == 'F' or efund_name == 'VS':
+                        self['IN3'][f'A{efund_name}'].append(a1)
+                        self['IN3'][f'A{efund_name}2'].append(a2)
 
-            self['IN3'][f'I{efund_name}COIL'] += 1
+                    if efund_name == 'F':
+                        self['IN3'][f'TURN{efund_name}C'].append(1.0)
+                    if efund_name == 'VS':
+                        self['IN3'][f'VSID'].append(nsum[efund_name])
+                        self['IN3'][f'IVESEL'] = 1
+                    elif efund_name != 'ACOIL':
+                        self['IN3'][f'{efund_name}CID'].append(nsum[efund_name])
 
-        self['MACHINEIN']['nfsum'] = self['machinein']['nfcoil'] = ncoil['F']
-        self['MACHINEIN']['nesum'] = self['machinein']['necoil'] = ncoil['E']
-        self['machinein']['nacoil'] = ncoil['A']
+                        self['IN3'][f'{efund_name}CTURN'].append(coil['element'][0]['turns_with_sign'])
+                        self['IN3'][f'I{efund_name}COIL'] = 1
+                    else:
+                        self['IN3'][f'I{efund_name}'] = 1
+
+        self['MACHINEIN']['nfsum'] = nsum['F']
+        self['machinein']['nfcoil'] = ncoil['F']
+        self['MACHINEIN']['nesum'] = nsum['E']
+        self['machinein']['necoil'] = ncoil['E']
+        self['MACHINEIN']['nvsum'] = nsum['VS']
+        self['machinein']['nvesel'] = ncoil['VS']
+
+        self['MACHINEIN']['nacoil'] = ncoil['ACOIL']
 
         if 'magnetics' in ods:
             if 'b_field_pol_probe' in ods['magnetics']:
-                self['MACHINEIN']['magpr2'] = len(ods['magnetics.b_field_pol_probe'])
-                self['IN3']['MPNAM2'] = ods[f'magnetics.b_field_pol_probe.:.name']
-                self['IN3']['XMP2'] = ods[f'magnetics.b_field_pol_probe.:.position.r']
-                self['IN3']['YMP2'] = ods[f'magnetics.b_field_pol_probe.:.position.z']
-                self['IN3']['SMP2'] = ods[f'magnetics.b_field_pol_probe.:.length']
-                self['IN3']['AMP2'] = -180 / np.pi * ods[f'magnetics.b_field_pol_probe.:.poloidal_angle']
+                self['MACHINEIN']['magpr2'] = nmp2 = len(list(filter(None, ods['magnetics.b_field_pol_probe.:.length'])))
+                self['IN3']['XMP2'] = xmp2 = np.zeros(nmp2)
+                self['IN3']['YMP2'] = ymp2 = np.zeros(nmp2)
+                self['IN3']['SMP2'] = smp2 = np.zeros(nmp2)
+                self['IN3']['AMP2'] = amp2 = np.zeros(nmp2)
+                self['IN3']['MPNAM2'] = mpnam2 = np.chararray(nmp2, itemsize=10, unicode=True)
+                n = 0
+                for i in ods['magnetics.b_field_pol_probe']:
+                    probe = ods['magnetics.b_field_pol_probe'][i]
+                    if 'length' in probe:
+                        xmp2[n] = probe['position.r']
+                        ymp2[n] = probe['position.z']
+                        smp2[n] = probe['length']
+                        amp2[n] = -180 / np.pi * probe['poloidal_angle']
+                        mpnam2[n] = probe['name']
+                        n += 1
 
             if 'flux_loop' in ods['magnetics']:
-                self['MACHINEIN']['nsilop'] = len(ods['magnetics.flux_loop'])
-                self['IN3']['LPNAME'] = ods[f'magnetics.flux_loop.:.name']
-                self['IN3']['RSI'] = ods[f'magnetics.flux_loop.:.position[0].r']
-                self['IN3']['ZSI'] = ods[f'magnetics.flux_loop.:.position[0].z']
+                self['MACHINEIN']['nsilop'] = nsilop = len(ods['magnetics.flux_loop'])
+                self['IN3']['RSI'] = rsi = np.zeros(nsilop)
+                self['IN3']['ZSI'] = zsi = np.zeros(nsilop)
+                self['IN3']['LPNAME'] = lpname = np.chararray(nsilop, itemsize=10, unicode=True)
+                for i in ods['magnetics.flux_loop']:
+                    loop = ods['magnetics.flux_loop'][i]
+                    # this assumes simple loops and
+                    # mean(psi(r1,z1), psi(r2,z2)) = psi(mean(r1, r2), mean(z1, z2))
+                    r = loop['position.0.r']
+                    z = loop['position.0.z']
+                    n = 1
+                    # last position in loop may be repeated
+                    for j in loop['position']:
+                        if loop['position'][j] == loop['position'][0]:
+                            continue
+                        r += loop['position'][j]['r']
+                        z += loop['position'][j]['z']
+                        n += 1
+                    rsi[i] = r / n
+                    zsi[i] = z / n
+                    lpname[i] = loop['name']
 
+        # NOTE: the vacuum vessel could also be described in 'pf_passive' (not yet included)
         if 'wall.description_2d[0].vessel.unit' in ods:
             self['IN3']['IVESEL'] = 1
 
-            nvesel = len(ods['wall.description_2d.0.vessel.unit'])
-            mdin['MACHINEIN']['nvesel'] = mdin['MACHINEIN']['nvsum'] = nvesel
+            nvesel = 0
+            for i in ods['wall.description_2d.0.vessel.unit']:
+                unit = ods['wall.description_2d.0.vessel.unit'][i]
+                if 'element' in unit:
+                    nvesel += 1
+                elif 'annular' in unit:
+                    nvesel += len(unit['annular.centreline.r']) - 1 + unit['annular.centreline.closed']
+                else:
+                    raise ValueError(f'No conversion defined for geometry of vessel unit {iunit}')
+            self['MACHINEIN']['nvesel'] = self['MACHINEIN']['nvsum'] = nvesel
             self['IN3']['RVS'] = rvs = np.zeros(nvesel)
             self['IN3']['ZVS'] = zvs = np.zeros(nvesel)
             self['IN3']['WVS'] = wvs = np.zeros(nvesel)
             self['IN3']['HVS'] = hvs = np.zeros(nvesel)
             self['IN3']['AVS'] = avs = np.zeros(nvesel)
             self['IN3']['AVS2'] = avs2 = np.zeros(nvesel)
-            for unit in ods['wall.description_2d.0.vessel.unit']:
-                rvs[i], zvs[i], wvs[i], hvs[i], avs[i], avs2[i] = outline_to_efund(
-                    ods[f'wall.description_2d.0.vessel.unit.{unit}.element.0.outline']
-                )
-                self['IN3']['RSISVS'] = ods[f'wall.description_2d.0.vessel.unit.{unit}.element.0.resistivity']
+            self['IN3']['RSISVS'] = rsisvs = np.zeros(nvesel)
+            self['IN3']['VSID'] = vsid = np.zeros(nvesel, dtype=int)
+            self['IN3']['VSNAME'] = vsname = np.chararray(nvesel, itemsize=10, unicode=True)
+            n = 0
+            for i in ods['wall.description_2d.0.vessel.unit']:
+                unit = ods['wall.description_2d.0.vessel.unit'][i]
+                if 'element' in unit:
+                    rvs[n], zvs[n], wvs[n], hvs[n], avs[n], avs2[n] = self.outline_to_efund(unit['element.0.outline'])
+                    rsisvs[n] = unit['element.0.resistivity']
+                    vsid[n] = n + 1
+                    vsname[n] = unit['name']
+                    n += 1
+                elif 'annular' in unit:
+                    nl = n + len(unit['annular.centreline.r']) - 1 + unit['annular.centreline.closed']
+                    rvs[n:nl], zvs[n:nl], wvs[n:nl], hvs[n:nl], avs[n:nl], avs2[n:nl] = self.annular_to_efund(unit['annular'])
+                    rsisvs[n:nl] = unit['annular.resistivity']
+                    vsid[n:nl] = np.arange(n + 1, nl + 1)
+                    vsname[n:nl] = unit['name']
+                    n = nl
 
-        if 'wall.description_2d[0].limiter.unit[0].outline' in ods:
-            rwall = ods['wall.description_2d[0].limiter.unit[0].outline.r']
-            zwall = ods['wall.description_2d[0].limiter.unit[0].outline.z']
+        # currently unused (not part of mhdin.dat, but could be used to make lim.dat files)
+        # if 'wall.description_2d[0].limiter.unit[0].outline' in ods:
+        #    rwall = ods['wall.description_2d[0].limiter.unit[0].outline.r']
+        #    zwall = ods['wall.description_2d[0].limiter.unit[0].outline.z']
 
         return self
 
@@ -1112,20 +1324,37 @@ class OMFITmhdin(OMFITnamelist):
                     ods[f'magnetics.b_field_pol_probe.{k}.turns'] = 1
 
         # Vessel
-        if 'vessel' in update and 'RVS' in self['IN3']:
-            rvs = self['IN3']['RVS']
-            zvs = self['IN3']['ZVS']
-            wvs = self['IN3']['WVS']
-            hvs = self['IN3']['HVS']
-            avs = self['IN3']['AVS']
-            avs2 = self['IN3']['AVS2']
-            for iunit, unit in enumerate(self['IN3']['VSID']):
-                self.efund_to_outline(
-                    [rvs[iunit], zvs[iunit], wvs[iunit], hvs[iunit], avs[iunit], avs2[iunit]],
-                    ods[f'wall.description_2d.0.vessel.unit.{iunit}.element.0.outline'],
-                )
+        if 'vessel' in update and 'RVS' in self['IN3'] and len(self['IN3']['RVS']):
+            r = self['IN3']['RVS']
+            z = self['IN3']['ZVS']
+            width = self['IN3']['WVS']
+            height = self['IN3']['HVS']
+            angle1 = self['IN3']['AVS']
+            angle2 = self['IN3']['AVS2']
+            resistivity = self['IN3']['RSISVS']
+            elements_id = (self['IN3']['VSID'] - 1).astype(int)
 
-                ods[f'wall.description_2d.0.vessel.unit.{iunit}.element.0.resistivity'] = self['IN3']['RSISVS'][iunit]
+            for i in range(len(r)):
+                c = elements_id[i]
+                e = sum(elements_id[:i] == elements_id[i])
+                ods['pf_passive.loop'][c]['name'] = f'VS{c}'
+                # ods['pf_passive.loop'][c]['identifier'] = f'PF{c}'
+                ods['pf_passive.loop'][c]['element'][e]['name'] = f'VS{c}_{e}'
+                ods['pf_passive.loop'][c]['element'][e]['identifier'] = f'VS{c}_{e}'
+                ods['pf_passive.loop'][c]['element'][e]['turns_with_sign'] = 1.0
+                ods['pf_passive.loop'][c]['resistivity'] = resistivity[c]
+                if angle1[i] == 0 and angle2[i] == 0:
+                    rect = ods['pf_passive.loop'][c]['element'][e]['geometry.rectangle']
+                    rect['r'] = r[i]
+                    rect['z'] = z[i]
+                    rect['width'] = width[i]
+                    rect['height'] = height[i]
+                    ods['pf_passive.loop'][c]['element'][e]['geometry.geometry_type'] = 2
+                else:
+                    outline = ods['pf_passive.loop'][c]['element'][e]['geometry.outline']
+                    outline = self.efund_to_outline([r[i], z[i], width[i], height[i], angle1[i], angle2[i]], outline)
+
+                    ods['pf_passive.loop'][c]['element'][e]['geometry.geometry_type'] = 1
 
         return ods
 
@@ -1416,6 +1645,153 @@ def get_mhdindat(
     return mhd
 
 
+def green_to_omas(
+    ods=None,
+    filedir='/fusion/projects/codes/efit/efitai/efit_support_files/DIII-D/green/168191/',
+    nw=129,
+    nh=129,
+    nsilop=44,
+    magpr2=76,
+    nesum=6,
+    nfsum=18,
+    nvsum=24,
+):
+
+    """
+    This function reads EFUND generate Green's function tables and put them into IMAS
+
+    :param ods: input ods to populate
+
+    :param filedir: directory which contains EFUND generated binary files
+
+    :param nw: number of horizontal grid points
+
+    :param nw: number of vertical grid points
+
+    :param magpr2: number of magnetic probes (will be overwritten if available in directory)
+
+    :param nsilop: number of flux loops (will be overwritten if available in directory)
+
+    :param nesum: number of e-coils (will be overwritten if available in directory)
+
+    :param nfsum: number of f-coils (will be overwritten if available in directory)
+
+    :param nvsum: number of vessel structures (will be overwritten if available in directory)
+
+    returns ods
+
+    """
+    from omas import omas_environment, ODS
+
+    try:
+        filename = '/mhdin.dat'
+        file = filedir + filename
+        mhdin = OMFITnamelist(file)
+        if 'machinein' in mhdin:
+            nfsum = mhdin['machinein']['nfsum']
+            nesum = mhdin['machinein']['nesum']
+            nvsum = mhdin['machinein']['nvsum']
+            magpr2 = mhdin['machinein']['magpr2']
+            nsilop = mhdin['machinein']['nsilop']
+    except Exception:
+        filename = '/dprobe.dat'  # pretend dprobe file is mhdin for coil names
+        file = filedir + filename
+        mhdin = OMFITnamelist(file)
+        printw('could not read array sizes from mhdin.dat')
+
+    if ods is None:
+        ods = ODS()
+
+    ods['em_coupling.code.name'] = 'EFUND'
+    ods['em_coupling.poloidal_probes'] = mhdin['in3']['mpnam2']
+    ods['em_coupling.flux_loops'] = mhdin['in3']['lpname']
+
+    green = {}
+    green['nesum'] = nesum
+    green['nsilop'] = nsilop
+    green['magpr2'] = magpr2
+    green['nfsum'] = nfsum
+    green['nvsum'] = nvsum
+    green['nw'] = nw
+    green['nh'] = nh
+
+    endian = '>'
+    # Response of poloidal field (F) coils on grid, and grid on itself.
+    filename = f'ec{nw}{nh}.ddd'
+    file = filedir + filename
+    f = scipy.io.FortranFile(file, 'r', f'{endian}i4')
+    [mw, mh] = f.read_ints(f'{endian}i4')
+    grid = f.read_reals(f'{endian}f8')
+    green['rgrid'] = grid[:mw]
+    green['zgrid'] = grid[mh:]
+    green['RR'], green['ZZ'] = np.meshgrid(green['rgrid'], green['zgrid'])
+
+    ggridfc = f.read_reals(f'{endian}f8').reshape([nfsum, mw * mh]).T
+    ods['em_coupling.mutual_grid_grid'] = f.read_reals(f'{endian}f8').reshape(mw, mh * mw).T
+
+    # Response of poloidal field coil on flux loops, magnetic probes, and grid
+    filename = f'rfcoil.ddd'
+    file = filedir + filename
+    f = scipy.io.FortranFile(file, 'r', f'{endian}i4')
+    rsilfc = f.read_reals(f'{endian}f8').reshape([nfsum, nsilop]).T
+    rmp2fc = f.read_reals(f'{endian}f8').reshape([nfsum, magpr2]).T
+
+    # Response of Ohmic heating coil on flux loops, magnetic probes, and grid
+    try:
+        filename = f're{nw}{nh}.ddd'
+        file = filedir + filename
+        f = scipy.io.FortranFile(file, 'r', f'{endian}i4')
+        rsilec = f.read_reals(f'{endian}f8').reshape([nesum, nsilop]).T
+        rmp2ec = f.read_reals(f'{endian}f8').reshape([nesum, magpr2]).T
+        gridec = f.read_reals(f'{endian}f8').reshape([nesum, mw * mh]).T
+        ods['em_coupling.mutual_loops_active'] = np.append(rsilfc, rsilec, axis=1)
+        ods['em_coupling.field_probes_active'] = np.append(rmp2fc, rmp2ec, axis=1)
+        ods['em_coupling.mutual_grid_active'] = np.append(ggridfc, gridec, axis=1)
+        no_ecoil = False
+    except Exception:
+        ods['em_coupling.mutual_loops_active'] = rsilfc
+        ods['em_coupling.field_probes_active'] = rmp2fc
+        ods['em_coupling.mutual_grid_active'] = ggridfc
+        no_ecoil = True
+        printw('Ohmic heating coil tables not available')
+
+    # Response of vessel on flux loops, probes, and grid
+    try:
+        filename = f'rv{nw}{nh}.ddd'
+        file = filedir + filename
+        f = scipy.io.FortranFile(file, 'r', f'{endian}i4')
+        ods['em_coupling.mutual_loops_passive'] = f.read_reals(f'{endian}f8').reshape([nvsum, nsilop]).T
+        ods['em_coupling.field_probes_passive'] = f.read_reals(f'{endian}f8').reshape([nvsum, magpr2]).T
+        ods['em_coupling.mutual_grid_passive'] = f.read_reals(f'{endian}f8').reshape([nvsum, mw * mh]).T
+
+    except Exception:
+        printw('Vessel not available')
+
+    # Response of coils on vessel
+    try:
+        gfcvs = f.read_reals(f'{endian}f8').reshape([nvsum, nfsum])
+        gecvs = f.read_reals(f'{endian}f8').reshape([nvsum, nesum])
+
+        if not no_ecoil:
+            ods['em_coupling.mutual_passive_active'] = np.append(gfcvs, gecvs, axis=1)
+        else:
+            ods['em_coupling.mutual_passive_active'] = gfcvs
+
+        ods['em_coupling.mutual_passive_passive'] = f.read_reals(f'{endian}f8').reshape([nvsum, nvsum])
+
+    except Exception:
+        printw('Vessel-coil mutual inductances not available')
+
+    # Response of magnetics flux loops and probes on grid
+    filename = f'ep{nw}{nh}.ddd'
+    file = filedir + filename
+    f = scipy.io.FortranFile(file, 'r', f'{endian}i4')
+    ods['em_coupling.mutual_loops_grid'] = f.read_reals(f'{endian}f8').reshape([mw * mh, nsilop]).T
+    ods['em_coupling.field_probes_grid'] = f.read_reals(f'{endian}f8').reshape([mw * mh, magpr2]).T
+    ods['em_coupling.code.parameters'] = green
+    return ods
+
+
 ############################################
 if '__main__' == __name__:
     from matplotlib import pyplot
