@@ -225,9 +225,9 @@ def get_definition_from_docstring(theobject, return_kw=False):
     Reads example code from the docstring of a class or function and uses it to define some variables
 
     :param theobject: class or function
-        The docstring should contain statements like '>>> the_variable = the_value'.
+        The docstring should contain statements like '>> the_variable = the_value'.
         They can be multi-line.
-        Each statement should be separated from others by lines not starting with >>> .
+        Each statement should be separated from others by lines not starting with >> .
 
     :param return_kw: bool
         Instead of evaluating expressions as written (e.g. var = ClassName(**kw)),
@@ -240,15 +240,15 @@ def get_definition_from_docstring(theobject, return_kw=False):
         Values are evaluated expressions taken from the right side of the first =.
     """
     fn_doc = theobject.__doc__.split('\n')
-    is_example = np.array([doc_line.strip().startswith('>>> ') for doc_line in fn_doc])
+    is_example = np.array([doc_line.strip().startswith('>> ') for doc_line in fn_doc])
     exmp_changes = np.append(0, np.diff(is_example.astype(int)))
     starts = np.where(exmp_changes == 1)[0]
     ends = np.where(exmp_changes == -1)[0]
     docstring_definitions = {}
     for start, end in zip(starts, ends):
-        var_name = fn_doc[start].split('=')[0].strip().split('>>> ')[1]
+        var_name = fn_doc[start].split('=')[0].strip().split('>> ')[1]
         def_start = '='.join(fn_doc[start].split('=')[1:]).strip()
-        def_rest = [d.split('>>> ')[1] for d in fn_doc[start + 1 : end]]
+        def_rest = [d.split('>> ')[1] for d in fn_doc[start + 1 : end]]
         definition = '\n'.join([def_start] + def_rest)
         if return_kw:
             # noinspection PyBroadException
@@ -1141,14 +1141,14 @@ class OMFITexperimentTable(OMFITgoogleSheet):
         This sample should work with the example/test sheet.
 
         If the sheet in question had ONLY DIII-D data, the same signal request could be accomplished via:
-        >>> simple_sample_signal_request = dict(column_name='density', factor=1e-13, tmin='t min', tmax='t max')
+        >> simple_sample_signal_request = dict(column_name='density', factor=1e-13, tmin='t min', tmax='t max')
         This sample won't work with the test sheet; it will fail on the EAST shots.
         We are exploiting the shortcut that we've used a valid pointname (valid for d3d at least)
         as the column header. If you want fancy column names, you have to specify TDIs.
         We are also relying on the defaults working for this case.
 
         The sample code in this docstring is interpreted and used by the regression test, so don't break it.
-        Separate different samples with non-example lines (that don't start with >>>)
+        Separate different samples with non-example lines (that don't start with >>)
 
         :param signals: list of dict-likes
             Each item in this list should be a dict-like that contains information needed to fetch & process data.
