@@ -310,16 +310,20 @@ class OMFITmaintree(OMFITproject):
             self.tmpSkel['MainSettings'].recursiveUpdate(namelist.NamelistFile(skelMainSettings), overwrite=True)
 
             # institution settings
-            filename = os.sep.join([OMFITsrc, '..', 'institution'])
-            if os.path.exists(filename):
-                tmp = namelist.NamelistFile(filename)
+            if os.path.exists(os.environ.get('OMFIT_INSTITUTION_FILE', '/does not exist')):
+                tmp = namelist.NamelistFile(os.environ['OMFIT_INSTITUTION_FILE'])
             else:
-                institution_files = glob.glob(os.sep.join([OMFITsrc, '..', 'institutions', '']) + '*')
-                institution_files.append(None)
-                for filename in institution_files:
+                filename = os.sep.join([OMFITsrc, '..', 'institution'])
+
+                if os.path.exists(filename):
                     tmp = namelist.NamelistFile(filename)
-                    if 'SETUP' in tmp and 'stats_file' in tmp['SETUP'] and os.path.exists(tmp['SETUP']['stats_file']):
-                        break
+                else:
+                    institution_files = glob.glob(os.sep.join([OMFITsrc, '..', 'institutions', '']) + '*')
+                    institution_files.append(None)
+                    for filename in institution_files:
+                        tmp = namelist.NamelistFile(filename)
+                        if 'SETUP' in tmp and 'stats_file' in tmp['SETUP'] and os.path.exists(tmp['SETUP']['stats_file']):
+                            break
             self.tmpSkel['MainSettings'].recursiveUpdate(tmp, overwrite=True)
 
             # user settings
